package com.smartfoxserver.v2.redbox;

import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.smartfoxserver.v2.core.SFSEventType;
import com.smartfoxserver.v2.exceptions.ExceptionMessageComposer;
import com.smartfoxserver.v2.extensions.SFSExtension;
import com.smartfoxserver.v2.redbox.handlers.AVChatHandler;
import com.smartfoxserver.v2.redbox.handlers.AVClipHandler;
import com.smartfoxserver.v2.redbox.handlers.ServerEventHandler;
import com.smartfoxserver.v2.redbox.managers.AVChatManager;
import com.smartfoxserver.v2.redbox.managers.AVClipManager;
import com.smartfoxserver.v2.redbox.utils.Constants;

/**
 * SmartFoxServer 2X RedBox extension.
 * This class is responsible for dispatching requests to sub-managers, like the AVClipManager or the AVChatManager.
 * 
 * @version	1.0.0
 * 
 * @author	The gotoAndPlay() Team
 * 			http://www.smartfoxserver.com
 * 			http://www.gotoandplay.it
 */
public class RedBoxExtension extends SFSExtension
{
	//--------------------------------------
	//  PRIVATE VARIABLES
	//--------------------------------------
	
	private static final String logPrefix = "[RedBox] ";
	private static Logger log = LoggerFactory.getLogger(RedBoxExtension.class);
	private static boolean logInfoEnabled;
	
	private Properties redBoxConfig;
	
	private AVClipManager avClipManager;
	private AVChatManager avChatManager;
	
	//-------------------------------------------------------
	// PUBLIC METHODS
	//-------------------------------------------------------
	
	/**
	 * Initialize RedBox Core Extension.
	 * The external properties file is loaded, the logger is configured and the requests managers are instantiated.
	 * 
	 * In case this method is overridden, it is mandatory to call the super() method.
	 */
	@Override
	public void init()
	{
		// Load RedBox external configuration
		loadConfig();
		
		if (redBoxConfig != null)
		{
			// Enable logging
			logInfoEnabled = (getRedBoxConfigParam("log").equals("1"));
			
			// Create AVClipManager and its requests and events handler
			if (getRedBoxConfigParam("enableAVClipManager").equals("1"))
			{
				avClipManager = new AVClipManager(this);
				addRequestHandler(Constants.AVCLIP_HANDLER_PREFIX, AVClipHandler.class);
				addEventHandler(SFSEventType.USER_LOGOUT, ServerEventHandler.class);
			}
			
			// Create AVChatManager and its requests and events handler
			if (getRedBoxConfigParam("enableAVChatManager").equals("1"))
			{
				avChatManager = new AVChatManager(this);
				addRequestHandler(Constants.AVCHAT_HANDLER_PREFIX, AVChatHandler.class);
				addEventHandler(SFSEventType.USER_DISCONNECT, ServerEventHandler.class);
			}
			
			logInfo("Core Extension initialization completed");
			
			if (avClipManager == null && avChatManager == null)
				logWarning("Both AVClipManager and AVChatManager are disabled: RedBoxExtension won't be able to handle client requests");
		}
		else
			logError("Core Extension initialization failed due to missing configuration");
	}
	
	//-------------------------------------------------------
	// PUBLIC FINAL METHODS
	//-------------------------------------------------------
	
	/**
	 * @exclude
	 */
	public final String getRedBoxConfigParam(String name)
	{
		return redBoxConfig.getProperty(name);
	}
	
	/**
	 * @exclude
	 */
	public final AVClipManager getAVClipManager()
	{
		return avClipManager;
	}
	
	/**
	 * @exclude
	 */
	public final AVChatManager getAVChatManager()
	{
		return avChatManager;
	}
	
	// LOGGING METHODS
	
	public static void logInfo(String info)
	{
		if (logInfoEnabled)
			log.info(logPrefix + info);
	}
	
	public static void logWarning(String warn)
	{
		log.warn(logPrefix + warn);
	}
	
	public static void logError(String err)
	{
		log.error(logPrefix + err);
	}
	
	public static void logFullException(ExceptionMessageComposer msg)
	{
		log.error(msg.toString());
	}
	
	//-------------------------------------------------------
	// PRIVATE METHODS
	//-------------------------------------------------------
	
	/**
	 * Load external configuration file.
	 */
	private void loadConfig()
	{
		redBoxConfig = new Properties();
		
		try
		{
			redBoxConfig.load(new FileInputStream(this.getCurrentFolder() + Constants.CONFIG_FILE));
			
			// Validate config params
			validateConfigParam("log", "0");
			validateConfigParam("enableAVClipManager", "0");
			validateConfigParam("enableAVChatManager", "0");
			validateConfigParam("red5Path", "");
			validateConfigParam("lastModifiedFormat", Constants.DEFAULT_LASTMODIFIED_FORMAT);
		}
		catch (IOException error)
		{
			redBoxConfig = null;
			logError("Unable to load '" + Constants.CONFIG_FILE + "' file containing RedBox configuration parameters");
		}
	}
	
	private void validateConfigParam(String paramName, String defaultValue)
	{
		if (getRedBoxConfigParam(paramName) == null)
		{
			logWarning("Configuration file doesn't contain '" + paramName + "' parameter; RedBox may not work properly");
			redBoxConfig.setProperty(paramName, defaultValue);
		}
	}
}
