package com.smartfoxserver.v2.redbox.managers;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.concurrent.ConcurrentHashMap;

import com.smartfoxserver.v2.core.ISFSEvent;
import com.smartfoxserver.v2.core.SFSEventParam;
import com.smartfoxserver.v2.core.SFSEventType;
import com.smartfoxserver.v2.entities.User;
import com.smartfoxserver.v2.entities.data.ISFSObject;
import com.smartfoxserver.v2.entities.data.SFSObject;
import com.smartfoxserver.v2.exceptions.SFSException;
import com.smartfoxserver.v2.extensions.SFSExtension;
import com.smartfoxserver.v2.redbox.RedBoxExtension;
import com.smartfoxserver.v2.redbox.data.ChatSession;
import com.smartfoxserver.v2.redbox.utils.Constants;

/**
 * SmartFoxServer 2X RedBox Audio/Video Chat Manager.
 * This class is responsible for audio/video chat sessions handling.
 * 
 * @version	1.0.0
 * 
 * @author	The gotoAndPlay() Team
 * 			http://www.smartfoxserver.com
 * 			http://www.gotoandplay.it
 */
public class AVChatManager extends BaseManager
{
	//--------------------------------------
	// CLASS CONSTANTS
	//--------------------------------------
	
	private static final String TIMESTAMP_FORMAT = "yyyyMMddHHmmssSSS";
	
	// Request type
	private static final String REQ_TYPE_SEND = "send";				// Stream from requester to recipient
	private static final String REQ_TYPE_RECEIVE = "receive";		// Stream from recipient to requester
	private static final String REQ_TYPE_SEND_RECEIVE = "send&rcv";	// Bi-diretional stream
	
	// Incoming commands
	private static final String CMD_REQUEST = "req";
	private static final String CMD_ACCEPT = "accept";
	private static final String CMD_REFUSE = "refuse";
	private static final String CMD_STOP = "stop";
	
	// Outgoing responses
	private static final String RES_REQUEST = "req";
	private static final String RES_START = "start";
	private static final String RES_REFUSED = "refused";
	private static final String RES_STOP = "stop";
	
	// Outgoing errors
	private static final String ERR_NO_RECIPIENT = "err_noRcp";
	private static final String ERR_DUPLICATE_REQUEST = "err_dup";
	
	//--------------------------------------
	//  PRIVATE VARIABLES
	//--------------------------------------
	
	private RedBoxExtension mainExtension;
	private ConcurrentHashMap<String, ChatSession> chatSessions;
	
	//-------------------------------------------------------
	// CONSTRUCTOR
	//-------------------------------------------------------
	
	/**
	 * The AVChatManager constructor.
	 */
	public AVChatManager(RedBoxExtension ext)
	{
		super(ext, Constants.AVCHAT_HANDLER_PREFIX);
		
		mainExtension = ext;
		
		chatSessions = new ConcurrentHashMap<String, ChatSession>();
		
		RedBoxExtension.logInfo("AVChatManager initialized");
	}
	
	//-------------------------------------------------------
	// PUBLIC METHODS
	//-------------------------------------------------------
	
	/**
	 * Handle client requests.
	 */
	public void handleClientRequest(User sender, ISFSObject params)
	{
		String cmd = params.getUtfString(SFSExtension.MULTIHANDLER_REQUEST_ID);
		
		// New chat request sent
		if (cmd.equals(CMD_REQUEST))
			handleChatRequest(sender, params);
		
		// Chat request refused
		else if (cmd.equals(CMD_REFUSE))
			handleChatRefused(sender, params);
		
		// Chat request accepted
		else if (cmd.equals(CMD_ACCEPT))
			handleChatAccepted(sender, params);
		
		// Chat stopped
		else if (cmd.equals(CMD_STOP))
			handleChatStopped(sender, params);
	}
	
	/**
	 * Handle internal events.
	 */
	public void handleServerEvent(ISFSEvent event) throws SFSException
	{
		SFSEventType evtType = event.getType();
		
		// User disconnected or logged out
		if (evtType == SFSEventType.USER_DISCONNECT || evtType == SFSEventType.USER_LOGOUT)
			handleRemoveUserRequest((User) event.getParameter(SFSEventParam.USER));
	}
	
	//-------------------------------------------------------
	// PRIVATE METHODS
	//-------------------------------------------------------
	
	// Request handlers -------------------------------------
	
	/**
	 * Validate request and forward to recipient.
	 */
	private void handleChatRequest(User requester, ISFSObject requestData)
	{
		int requesterId = requester.getId();
		String requesterName = requester.getName();
		
		int recipientId = requestData.getInt("uId");
		
		String requestId = requestData.getUtfString("id");
		String requestType = requestData.getUtfString("type");
		boolean enableCamera = requestData.getBool("cam");
		boolean enableMicrophone = requestData.getBool("mic");
		
		RedBoxExtension.logInfo("A/v chat request of type '" + requestType + "' sent by user " + requesterName + " to user id " + recipientId);
		
		// Validate request type
		boolean validType = (requestType.equals(REQ_TYPE_SEND) || requestType.equals(REQ_TYPE_RECEIVE) || requestType.equals(REQ_TYPE_SEND_RECEIVE));
		
		if (validType)
		{
			// Validate request id
			String validId = requestType + "-" + String.valueOf(requesterId) + "-" + String.valueOf(recipientId);
			
			if (requestId.equals(validId))
			{
				// Validate request
				boolean validRequest = true;
				
				// Check if the same request has already been submitted
				if (chatSessions.get(requestId) != null)
				{
					validRequest = false;
					RedBoxExtension.logInfo("Request already sent before by the same requester: request discarded");
				}
				
				// Check if a mutual request has already been submitted
				String mutualReqId = "";
				
				if (requestType.equals(REQ_TYPE_SEND_RECEIVE))
				{
					// Mutual request has same type but inverse requester/recipient
					mutualReqId = requestType + "-" + String.valueOf(recipientId) + "-" + String.valueOf(requesterId);
				}
				else if (requestType.equals(REQ_TYPE_SEND))
				{
					// Mutual request has "RECEIVE" type and inverse requester/recipient
					mutualReqId = REQ_TYPE_RECEIVE + "-" + String.valueOf(recipientId) + "-" + String.valueOf(requesterId);
				}
				else if (requestType.equals(REQ_TYPE_RECEIVE))
				{
					// Mutual request has "SEND" type and inverse requester/recipient
					mutualReqId = REQ_TYPE_SEND + "-" + String.valueOf(recipientId) + "-" + String.valueOf(requesterId);
				}
				
				if (chatSessions.get(mutualReqId) != null)
				{
					validRequest = false;
					RedBoxExtension.logInfo("Mutual request already sent before by the recipient: request discarded");
				}
				
				if (validRequest)
				{
					// Check if recipient is still available: if yes, send request to recipient; if no, send error to requester
					User recipient = mainExtension.getParentZone().getUserById(recipientId);
					
					if (recipient != null)
					{
						// Create a new session
						ChatSession session = new ChatSession();
						session.id = requestId;
						session.type = requestType;
						session.pending = true;
						session.requesterId = requesterId;
						session.requesterName = requesterName;
						session.recipientId = recipientId;
						session.recipientName = recipient.getName();
						session.enableCam = enableCamera;
						session.enableMic = enableMicrophone;
						
						// Add to sessions manager
						chatSessions.put(session.id, session);
						
						// Send request to recipient
						ISFSObject params = new SFSObject();
						params.putUtfString("id", session.id);
						params.putUtfString("type", session.type);
						params.putInt("uId", session.requesterId);
						params.putUtfString("uName", session.requesterName);
						params.putBool("cam", session.enableCam);
						params.putBool("mic", session.enableMic);
						
						sendResponse(recipient, RES_REQUEST, params);
						
						RedBoxExtension.logInfo("Request sent to recipient; session id: " + session.id);
					}
					else
					{
						// Send "no recipient" error to requester
						ISFSObject params = new SFSObject();
						params.putUtfString("id", requestId);
						
						sendResponse(requester, ERR_NO_RECIPIENT, params);
						
						RedBoxExtension.logInfo("Recipient not available: error sent to requester");
					}
				}
				else
				{
					// Send "duplicate request" error to requester
					ISFSObject params = new SFSObject();
					params.putUtfString("id", requestId);
					
					sendResponse(requester, ERR_DUPLICATE_REQUEST, params);
					
					RedBoxExtension.logInfo("Duplicate request error sent to requester");
				}
			}
			else
				RedBoxExtension.logWarning("Invalid a/v chat request id: '" + requestId + "'; request discarded");
		}
		else
			RedBoxExtension.logWarning("Invalid a/v chat request type: '" + requestType + "'; request discarded");
	}
	
	/**
	 * Handle refusal and forward to requester.
	 */
	private void handleChatRefused(User recipient, ISFSObject responseData)
	{
		String sessionId = responseData.getUtfString("id");
		
		RedBoxExtension.logInfo("A/v chat request '" + sessionId + "' refused by user " + recipient.getName() + " (" + recipient.getId() + ")");
		
		// Retrieve session
		ChatSession session = chatSessions.get(sessionId);
		
		// Check if session still exists
		if (session != null && session.pending)
		{
			// Check if requester is still available
			User requester = mainExtension.getParentZone().getUserById(session.requesterId);
			
			if (requester != null)
			{
				// Send "connection refused" event to requester
				ISFSObject params = new SFSObject();
				params.putUtfString("id", session.id);
				params.putUtfString("uName", session.recipientName);
				
				sendResponse(requester, RES_REFUSED, params);
			}
		
			// Remove session
			chatSessions.remove(sessionId);
		}
	}
	
	/**
	 * Handle acceptance: start chat on both clients.
	 */
	private void handleChatAccepted(User recipient, ISFSObject responseData)
	{
		String sessionId = responseData.getUtfString("id");
		
		RedBoxExtension.logInfo("A/v chat request '" + sessionId + "' accepted by user " + recipient.getName() + " (" + recipient.getId() + ")");
		
		// Retrieve session
		ChatSession session = chatSessions.get(sessionId);
		
		// Check if session still exists
		if (session != null && session.pending)
		{
			// Check if requester is still available
			User requester = mainExtension.getParentZone().getUserById(session.requesterId);
			
			if (requester != null)
			{
				String zone = mainExtension.getParentZone().getName();
				SimpleDateFormat dateFormatter = new SimpleDateFormat(TIMESTAMP_FORMAT);
				String timeStamp = dateFormatter.format(new Date());
				
				// Generate requester's stream id
				String requesterStreamId = null;
				if (session.type.equals(REQ_TYPE_SEND) || session.type.equals(REQ_TYPE_SEND_RECEIVE))
					requesterStreamId = zone + "_" + timeStamp + "_" + requester.getId();
				
				// Generate recipient's stream id
				String recipientStreamId = null;
				if (session.type.equals(REQ_TYPE_RECEIVE) || session.type.equals(REQ_TYPE_SEND_RECEIVE))
					recipientStreamId = zone + "_" + timeStamp + "_" + recipient.getId();
				
				// Update chat session data
				session.pending = false;
				session.requesterStreamId = requesterStreamId;
				session.recipientStreamId = recipientStreamId;
				
				// Send "start chat" event to both requester and recipient
				sendStartChatResponse(session.id, requester, requesterStreamId, recipient.getName(), recipientStreamId);
				sendStartChatResponse(session.id, recipient, recipientStreamId, requester.getName(), requesterStreamId);
			}
		}
	}
	
	/**
	 * Handle chat stop.
	 */
	private void handleChatStopped(User sender, ISFSObject responseData)
	{
		String sessionId = responseData.getUtfString("id");
		
		RedBoxExtension.logInfo("A/v chat request '" + sessionId + "' stopped by user " + sender.getName() + " (" + sender.getId() + ")");
		
		// Retrieve session
		ChatSession session = chatSessions.get(sessionId);
		
		// Check if session still exists
		if (session != null)
		{
			// Check if mate is still available
			int toUserId = (sender.getId() == session.requesterId ? session.recipientId : session.requesterId);
			User toUser = mainExtension.getParentZone().getUserById(toUserId);
			
			if (toUser != null)
			{
				// Send "stop chat" event to mate
				sendStopChatResponse(toUser, session);
				
				// Remove session
				chatSessions.remove(session.id);
			}
		}
	}
	
	/**
	 * Stop all session in which user is involved.
	 */
	private void handleRemoveUserRequest(User user)
	{
		for (ChatSession session : chatSessions.values())
		{
			// Check if user is involved in session
			if (session.requesterId == user.getId() || session.recipientId == user.getId())
			{
				
				ISFSObject obj = new SFSObject();
				obj.putUtfString("id", session.id);
				
				// Simulate external stop command
				handleChatStopped(user, obj);
			}
		}
	}
	
	// Other private methods --------------------------------
	
	/**
	 * Send "start chat" response to clients.
	 */
	private void sendStartChatResponse(String sessionId, User user, String userStream, String mateName, String mateStream)
	{
		ISFSObject params = new SFSObject();
		params.putUtfString("id", sessionId);
		params.putUtfString("mName", mateName);
		
		if (userStream != null)
			params.putUtfString("stream", userStream);
		
		if (mateStream != null)
			params.putUtfString("mStream", mateStream);
		
		sendResponse(user, RES_START, params);
	}
	
	/**
	 * Send "stop chat" response to client.
	 */
	private void sendStopChatResponse(User toUser, ChatSession session)
	{
		ISFSObject params = new SFSObject();
		params.putUtfString("id", session.id);
		
		// We add the mate name to the response in case the session was stopped by the recipient just after he received the request,
		// but before he was able to accept it (due to disconnection for example)
		String mateName = (toUser.getId() == session.requesterId ? session.recipientName : session.requesterName);
		params.putUtfString("mName", mateName);
		
		sendResponse(toUser, RES_STOP, params);
	}
}
